package de.akquinet.zalit.server.util;

import java.text.DecimalFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.Locale;

/**
 * @copyright akquinet AG, 2002-2010
 */
public final class DateUtil {
  private static final SimpleDateFormat DEUTSCHE_NOTATION_FORMAT_OHNE_JAHR = new SimpleDateFormat("dd.MM");
  private static final SimpleDateFormat MONAT_JAHR = new SimpleDateFormat("MM/yyyy");
  private static final SimpleDateFormat DEUTSCHE_NOTATION_FORMAT = new SimpleDateFormat("dd.MM.yyyy");
  private static final SimpleDateFormat DEUTSCHE_NOTATION_FORMAT_MIT_UHRZEIT = new SimpleDateFormat("dd.MM.yyyy HH:mm");

  private static final SimpleDateFormat TAG_FOTMAT_ZWEISTELLIG = new SimpleDateFormat("EE");

  private static final SimpleDateFormat DEUTSCHE_NOTATION_FORMAT_MIT_WOCHENTAG = new SimpleDateFormat("EEEEEEEEEEE, dd.MM.yyyy");

  private static final SimpleDateFormat WOCHENTAG = new SimpleDateFormat("EEEEEEEEEEE");

  private static final SimpleDateFormat UHRZEIT = new SimpleDateFormat("HH:mm");
  private static final DecimalFormat WOCHEN_FORMAT = new DecimalFormat("00");

  private static final int EIN_TAG_IN_MS = 1000 * 60 * 60 * 24; // 86400000
  private static final int ZEITUMSTELLUNGSSTUNDE = 3600000;

  public static final Date ARMAGEDDON = new DateUtil(1, 1, 2500).build();
  public static final Date URKNALL = new DateUtil(1, 1, 2000).build();

  private final Calendar _calendar = Calendar.getInstance();

  public DateUtil() {
    this(new Date());
  }

  public DateUtil(final Date date) {
    setDate(date);
  }

  public DateUtil(final Calendar calendar) {
    this(calendar.getTime());
  }

  public DateUtil(final String dateInGermanNotation) {
    try {
      final Date date;
      synchronized (DEUTSCHE_NOTATION_FORMAT) {
        date = DEUTSCHE_NOTATION_FORMAT.parse(dateInGermanNotation);
      }
      setDate(date);
    } catch (final ParseException e) {
      throw new IllegalArgumentException(dateInGermanNotation, e);
    }
  }

  public DateUtil(final int day, final int month, final int year, final int hour, final int minute, final int seconds) {
    setDate(day, month, year);
    setTime(hour, minute, seconds);
  }

  public DateUtil(final int day, final int month, final int year) {
    this(day, month, year, 0, 0, 0);
  }

  public String formatDate() {
    synchronized (DEUTSCHE_NOTATION_FORMAT) {
      return DEUTSCHE_NOTATION_FORMAT.format(build());
    }
  }

  public String formatDateAndTime() {
    synchronized (DEUTSCHE_NOTATION_FORMAT_MIT_UHRZEIT) {
      return DEUTSCHE_NOTATION_FORMAT_MIT_UHRZEIT.format(build());
    }
  }

  public String formatTime() {
    synchronized (UHRZEIT) {
      return UHRZEIT.format(build());
    }
  }

  public String formatDateAndDayOfWeek() {
    synchronized (DEUTSCHE_NOTATION_FORMAT_MIT_WOCHENTAG) {
      return DEUTSCHE_NOTATION_FORMAT_MIT_WOCHENTAG.format(build());
    }
  }

  public int getDay() {
    return _calendar.get(Calendar.DAY_OF_MONTH);
  }

  public DateUtil startOfYear() {
    _calendar.set(Calendar.DAY_OF_YEAR, 1);
    return this;
  }

  public int getDayOfWeek() {
    return _calendar.get(Calendar.DAY_OF_WEEK);
  }

  public int getHour() {
    return _calendar.get(Calendar.HOUR_OF_DAY);
  }

  public int getMinute() {
    return _calendar.get(Calendar.MINUTE);
  }

  public int getSecond() {
    return _calendar.get(Calendar.SECOND);
  }

  public int getMonth() {
    return _calendar.get(Calendar.MONTH);
  }

  public int getYear() {
    return _calendar.get(Calendar.YEAR);
  }

  public DateUtil shiftedByDays(final int days) {
    _calendar.add(Calendar.DAY_OF_YEAR, days);
    return this;
  }

  public static int periodInDays(final Date fromDate, final Date toDate, final boolean fiscalYear) {
    Date day1;
    Date day2;
    int result = 0;

    if (fromDate.compareTo(toDate) < 0) {
      day1 = fromDate;
      day2 = toDate;
    } else if (fromDate.compareTo(toDate) > 0) {
      day1 = toDate;
      day2 = fromDate;
    } else {
      return 1;
    }

    if (fiscalYear) {
      result = calcFiscalYear(day1, day2);
    } else {
      result = calcNormalYear(day1, day2);
    }

    return result;
  }

  public static Double periodInHour(final Date vonDatum, final Date bisDatum) {
    assert vonDatum.compareTo(bisDatum) < 0 : "VonDatum grsesser BisDatum";

    final long von = vonDatum.getTime();
    final long bis = bisDatum.getTime();
    final double differenz = bis - von;

    final Double result = new Double(differenz / 3600000);

    return result;
  }

  private static int calcNormalYear(final Date day1, final Date day2) {
    int result = 0;
    final GregorianCalendar cal = (GregorianCalendar) Calendar.getInstance();
    cal.setTime(day1);
    final int firstDays = cal.get(Calendar.DAY_OF_YEAR);
    int firstYear = cal.get(Calendar.YEAR);
    cal.setTime(day2);
    final int secondDays = cal.get(Calendar.DAY_OF_YEAR);
    final int secondYear = cal.get(Calendar.YEAR);
    while (secondYear > firstYear) {
      if (cal.isLeapYear(firstYear) && (firstYear != secondYear)) {
        result += 366;
      } else {
        result += 365;
      }
      firstYear++;
    }
    result += (secondDays - firstDays) + 1;
    return result;
  }

  private static int calcFiscalYear(final Date day1, final Date day2) {
    int result = 0;
    final Calendar cal = Calendar.getInstance();
    cal.setTime(day1);
    int firstDay = cal.get(Calendar.DAY_OF_MONTH);
    if (firstDay == 31) // 31. of each month is treated as the 30.
    {
      firstDay = 30;
    }
    int firstMonth = cal.get(Calendar.MONTH);
    int firstYear = cal.get(Calendar.YEAR);
    cal.setTime(day2);
    int secondDay = cal.get(Calendar.DAY_OF_MONTH);
    if (secondDay == 31) {
      secondDay = 30;
    }
    final int secondMonth = cal.get(Calendar.MONTH);
    final int secondYear = cal.get(Calendar.YEAR);
    while (secondYear != firstYear || secondMonth != firstMonth) {
      firstMonth++;
      if (firstMonth > 11) // Month start at 0!
      {
        firstMonth = 0;
        firstYear++;
      }
      result += 30;
    }
    result += (secondDay - firstDay) + 1;
    return result;
  }

  public DateUtil withoutTime() {
    return setTime(0, 0, 0, 0);
  }

  public DateUtil setDate(final Date date) {
    _calendar.setTime(date);
    return this;
  }

  public DateUtil setDate(final int day, final int month, final int year) {
    _calendar.set(year, month, day);
    return this;
  }

  public DateUtil setTime(final int hour) {
    return setTime(hour, 0);
  }

  public DateUtil setTime(final int hour, final int minute) {
    return setTime(hour, minute, 0);
  }

  public DateUtil setTime(final int hour, final int minute, final int second) {
    return setTime(hour, minute, second, 0);
  }

  public DateUtil setTime(final int hour, final int minute, final int second, final int milis) {
    _calendar.set(Calendar.HOUR_OF_DAY, hour);
    _calendar.set(Calendar.MINUTE, minute);
    _calendar.set(Calendar.SECOND, second);
    _calendar.set(Calendar.MILLISECOND, milis);
    return this;
  }

  public Date build() {
    return _calendar.getTime();
  }

  public int getWeekOfYear() {
    return _calendar.get(Calendar.WEEK_OF_YEAR);
  }

  public String getDayFormated() {
    synchronized (TAG_FOTMAT_ZWEISTELLIG) {
      return TAG_FOTMAT_ZWEISTELLIG.format(build());
    }
  }

  public List<Date> getAllCalendarWeeks() {
    withoutTime();
    final int year = getYear();
    final List<Date> alleKalenderWochenDerTageEinesJahres = new ArrayList<Date>();

    _calendar.set(Calendar.DAY_OF_YEAR, 1);

    while (_calendar.get(Calendar.DAY_OF_WEEK) != Calendar.MONDAY) {
      _calendar.add(Calendar.DATE, -1);
    }

    Date datum = null;

    do {
      datum = build();

      alleKalenderWochenDerTageEinesJahres.add(datum);

      shiftedByDays(7);
    } while (getYear() < year + 1);

    return alleKalenderWochenDerTageEinesJahres;
  }

  public String formatDateWithoutYear() {
    synchronized (DEUTSCHE_NOTATION_FORMAT_OHNE_JAHR) {
      return DEUTSCHE_NOTATION_FORMAT_OHNE_JAHR.format(build());
    }
  }

  public String formatMonthAndYear() {
    synchronized (MONAT_JAHR) {
      return MONAT_JAHR.format(build());
    }
  }

  public static int dateDiff(final Date einDatum, final Date anderesDatum) {
    Date day1;
    Date day2;

    if (einDatum.compareTo(anderesDatum) < 0) {
      day1 = einDatum;
      day2 = anderesDatum;
    } else if (einDatum.compareTo(anderesDatum) > 0) {
      day1 = anderesDatum;
      day2 = einDatum;
    } else {
      return 0;
    }

    final Calendar cal = Calendar.getInstance(Locale.GERMANY);
    cal.setTime(day1);
    final long firstInMs = cal.getTimeInMillis();
    cal.setTime(day2);
    final long secondInMs = cal.getTimeInMillis();

    int i = (int) ((secondInMs - firstInMs + ZEITUMSTELLUNGSSTUNDE) / EIN_TAG_IN_MS);

    if (einDatum.after(anderesDatum)) {
      i = i * -1;
    }
    return i;
  }

  public DateUtil endOfYear() {
    _calendar.set(Calendar.MONTH, Calendar.DECEMBER);
    _calendar.set(Calendar.DATE, 31);
    return this;
  }

  public String formatCalendarWeek() {
    synchronized (WOCHEN_FORMAT) {
      return WOCHEN_FORMAT.format(getWeekOfYear());
    }
  }

  public DateUtil shiftedByYears(final int shift) {
    _calendar.add(Calendar.YEAR, shift);
    return this;
  }

  public DateUtil shiftedByMonth(final int shift) {
    _calendar.add(Calendar.MONTH, shift);
    return this;
  }

  public DateUtil shiftedByMinutes(final int shift) {
    _calendar.add(Calendar.MINUTE, shift);
    return this;
  }

  public int getMaxDayOfMonth() {
    return _calendar.getActualMaximum(Calendar.DAY_OF_MONTH);
  }

  public List<Date> getVergangeneMonate(final int anzahl) {
    withoutDay();

    final List<Date> monate = new ArrayList<Date>();

    monate.add(build());
    for (int i = 0; i < anzahl - 1; i++) {
      monate.add(shiftedByMonth(-1).build());
    }
    return monate;
  }

  public DateUtil withoutDay() {
    withoutTime();
    return setDate(1, getMonth(), getYear());
  }

  public static String tag(final int day) {
    assert day >= 0 : "day >= 0";
    assert day <= 6 : "day < 6";
    final Calendar gregorianCalendar = new GregorianCalendar();
    gregorianCalendar.set(Calendar.DAY_OF_WEEK, day + 2);
    return WOCHENTAG.format(gregorianCalendar.getTime());
  }

  public Date getFirstDayOfWeek() {
    _calendar.set(Calendar.DAY_OF_WEEK, Calendar.MONDAY);
    return _calendar.getTime();
  }

}
